# coding: utf-8

"""
    Thoth User API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: 0.7.0-dev

    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class UnitDef(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'configuration': 'object',
        'name': 'str',
        'unit_run': 'bool'
    }

    attribute_map = {
        'configuration': 'configuration',
        'name': 'name',
        'unit_run': 'unit_run'
    }

    def __init__(self, configuration=None, name=None, unit_run=None):  # noqa: E501
        """UnitDef - a model defined in Swagger"""  # noqa: E501
        self._configuration = None
        self._name = None
        self._unit_run = None
        self.discriminator = None
        self.configuration = configuration
        self.name = name
        self.unit_run = unit_run

    @property
    def configuration(self):
        """Gets the configuration of this UnitDef.  # noqa: E501


        :return: The configuration of this UnitDef.  # noqa: E501
        :rtype: object
        """
        return self._configuration

    @configuration.setter
    def configuration(self, configuration):
        """Sets the configuration of this UnitDef.


        :param configuration: The configuration of this UnitDef.  # noqa: E501
        :type: object
        """
        if configuration is None:
            raise ValueError("Invalid value for `configuration`, must not be `None`")  # noqa: E501

        self._configuration = configuration

    @property
    def name(self):
        """Gets the name of this UnitDef.  # noqa: E501


        :return: The name of this UnitDef.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this UnitDef.


        :param name: The name of this UnitDef.  # noqa: E501
        :type: str
        """
        if name is None:
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def unit_run(self):
        """Gets the unit_run of this UnitDef.  # noqa: E501

        Flag set to true if the pipeline unit was run during the resolution process   # noqa: E501

        :return: The unit_run of this UnitDef.  # noqa: E501
        :rtype: bool
        """
        return self._unit_run

    @unit_run.setter
    def unit_run(self, unit_run):
        """Sets the unit_run of this UnitDef.

        Flag set to true if the pipeline unit was run during the resolution process   # noqa: E501

        :param unit_run: The unit_run of this UnitDef.  # noqa: E501
        :type: bool
        """
        if unit_run is None:
            raise ValueError("Invalid value for `unit_run`, must not be `None`")  # noqa: E501

        self._unit_run = unit_run

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(UnitDef, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, UnitDef):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other
