# coding: utf-8

"""
    Thoth User API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: 0.7.0-dev

    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class AdviserResultResponseResultReportDependencyGraph(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'edges': 'list[list[int]]',
        'nodes': 'list[str]'
    }

    attribute_map = {
        'edges': 'edges',
        'nodes': 'nodes'
    }

    def __init__(self, edges=None, nodes=None):  # noqa: E501
        """AdviserResultResponseResultReportDependencyGraph - a model defined in Swagger"""  # noqa: E501
        self._edges = None
        self._nodes = None
        self.discriminator = None
        self.edges = edges
        self.nodes = nodes

    @property
    def edges(self):
        """Gets the edges of this AdviserResultResponseResultReportDependencyGraph.  # noqa: E501

        Edge represented as a source-destination pair, representing indexes to the nodes listing   # noqa: E501

        :return: The edges of this AdviserResultResponseResultReportDependencyGraph.  # noqa: E501
        :rtype: list[list[int]]
        """
        return self._edges

    @edges.setter
    def edges(self, edges):
        """Sets the edges of this AdviserResultResponseResultReportDependencyGraph.

        Edge represented as a source-destination pair, representing indexes to the nodes listing   # noqa: E501

        :param edges: The edges of this AdviserResultResponseResultReportDependencyGraph.  # noqa: E501
        :type: list[list[int]]
        """
        if edges is None:
            raise ValueError("Invalid value for `edges`, must not be `None`")  # noqa: E501

        self._edges = edges

    @property
    def nodes(self):
        """Gets the nodes of this AdviserResultResponseResultReportDependencyGraph.  # noqa: E501

        A listing of nodes (package names) in the dependency graph   # noqa: E501

        :return: The nodes of this AdviserResultResponseResultReportDependencyGraph.  # noqa: E501
        :rtype: list[str]
        """
        return self._nodes

    @nodes.setter
    def nodes(self, nodes):
        """Sets the nodes of this AdviserResultResponseResultReportDependencyGraph.

        A listing of nodes (package names) in the dependency graph   # noqa: E501

        :param nodes: The nodes of this AdviserResultResponseResultReportDependencyGraph.  # noqa: E501
        :type: list[str]
        """
        if nodes is None:
            raise ValueError("Invalid value for `nodes`, must not be `None`")  # noqa: E501

        self._nodes = nodes

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(AdviserResultResponseResultReportDependencyGraph, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, AdviserResultResponseResultReportDependencyGraph):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other
