# coding: utf-8

"""
    Thoth User API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: 0.7.0-dev

    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class CallbackInfo(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'authorization': 'str',
        'client_data': 'object',
        'url': 'str'
    }

    attribute_map = {
        'authorization': 'authorization',
        'client_data': 'client_data',
        'url': 'url'
    }

    def __init__(self, authorization=None, client_data=None, url=None):  # noqa: E501
        """CallbackInfo - a model defined in Swagger"""  # noqa: E501
        self._authorization = None
        self._client_data = None
        self._url = None
        self.discriminator = None
        if authorization is not None:
            self.authorization = authorization
        if client_data is not None:
            self.client_data = client_data
        self.url = url

    @property
    def authorization(self):
        """Gets the authorization of this CallbackInfo.  # noqa: E501

        string to be passed as auth header to client webhook url  # noqa: E501

        :return: The authorization of this CallbackInfo.  # noqa: E501
        :rtype: str
        """
        return self._authorization

    @authorization.setter
    def authorization(self, authorization):
        """Sets the authorization of this CallbackInfo.

        string to be passed as auth header to client webhook url  # noqa: E501

        :param authorization: The authorization of this CallbackInfo.  # noqa: E501
        :type: str
        """

        self._authorization = authorization

    @property
    def client_data(self):
        """Gets the client_data of this CallbackInfo.  # noqa: E501

        arbitrary json that is passed back to the user upon request completion  # noqa: E501

        :return: The client_data of this CallbackInfo.  # noqa: E501
        :rtype: object
        """
        return self._client_data

    @client_data.setter
    def client_data(self, client_data):
        """Sets the client_data of this CallbackInfo.

        arbitrary json that is passed back to the user upon request completion  # noqa: E501

        :param client_data: The client_data of this CallbackInfo.  # noqa: E501
        :type: object
        """

        self._client_data = client_data

    @property
    def url(self):
        """Gets the url of this CallbackInfo.  # noqa: E501

        url to post results to upon request completion  # noqa: E501

        :return: The url of this CallbackInfo.  # noqa: E501
        :rtype: str
        """
        return self._url

    @url.setter
    def url(self, url):
        """Sets the url of this CallbackInfo.

        url to post results to upon request completion  # noqa: E501

        :param url: The url of this CallbackInfo.  # noqa: E501
        :type: str
        """
        if url is None:
            raise ValueError("Invalid value for `url`, must not be `None`")  # noqa: E501

        self._url = url

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(CallbackInfo, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CallbackInfo):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other
